import os
import base64
import codecs

def encode_pixels(pixels):
	from .datafile import Token
	return Token(codecs.decode(b"="+base64.b64encode(bytes(pixels))+b"=", "ascii"))

def pixels_to_rgba(pixels):
	return (int(p*255) for p in pixels)

def pixels_to_rgb(pixels):
	for i in range(0, len(pixels), 4):
		yield int(pixels[i]*255)
		yield int(pixels[i+1]*255)
		yield int(pixels[i+2]*255)

def pixels_to_rgb_invert_green(pixels):
	for i in range(0, len(pixels), 4):
		yield int(pixels[i]*255)
		yield 255-int(pixels[i+1]*255)
		yield int(pixels[i+2]*255)

def pixels_to_gray(pixels):
	for i in range(0, len(pixels), 4):
		yield int((pixels[i]+pixels[i+1]+pixels[i+2])*255/3)

class TextureExporter:
	def __init__(self):
		self.inline_data = True

	def export_texture(self, tex_node, usage='RGB', *, invert_green=False):
		image = tex_node.image
		from .datafile import Resource, Statement, Token
		tex_res = Resource(image.name+".tex2d", "texture2d")

		if tex_node.use_mipmap:
			tex_res.statements.append(Statement("generate_mipmap", True))

		colorspace = image.colorspace_settings.name
		if usage=='GRAY' and colorspace=='sRGB':
				raise Exception("Grayscale textures with sRGB colorspace are not supported")

		from .util import basename
		fn = basename(image.filepath)
		if not self.inline_data and not invert_green and fn:
			srgb = "_srgb" if colorspace=='sRGB' else ""
			tex_res.statements.append(Statement("external_image"+srgb, fn))
		else:
			if usage=='RGBA':
				fmt = 'SRGB8_ALPHA8' if colorspace=='sRGB' else 'RGBA8'
			elif usage=='GRAY':
				fmt = 'LUMINANCE8'
			else:
				fmt = 'SRGB8' if colorspace=='sRGB' else 'RGB8'

			tex_res.statements.append(Statement("storage", Token(fmt), image.size[0], image.size[1]))

			pixels = tuple(image.pixels)
			texdata = ""
			if usage=='RGBA':
				texdata = encode_pixels(pixels_to_rgba(pixels))
			elif usage=='GRAY':
				texdata = encode_pixels(pixels_to_gray(pixels))
			elif invert_green:
				texdata = encode_pixels(pixels_to_rgb_invert_green(pixels))
			else:
				texdata = encode_pixels(pixels_to_rgb(pixels))
			tex_res.statements.append(Statement("raw_data", texdata))

		return tex_res

class SamplerExporter:
	def export_sampler(self, tex_node):
		from .datafile import Resource, Statement, Token
		samp_res = Resource(self.get_sampler_name(tex_node), "sampler")

		use_interpolation = tex_node.interpolation!='Closest'
		if use_interpolation:
			if tex_node.use_mipmap:
				samp_res.statements.append(Statement("filter", Token('LINEAR_MIPMAP_LINEAR')))
			else:
				samp_res.statements.append(Statement("filter", Token('LINEAR')))
			samp_res.statements.append(Statement("max_anisotropy", tex_node.max_anisotropy))
		else:
			if tex_node.use_mipmap:
				samp_res.statements.append(Statement("filter", Token('NEAREST_MIPMAP_NEAREST')))
			else:
				samp_res.statements.append(Statement("filter", Token('NEAREST')))

		if tex_node.extension=="REPEAT":
			samp_res.statements.append(Statement("wrap", Token('REPEAT')))
		elif tex_node.extension=="EXTEND":
			samp_res.statements.append(Statement("wrap", Token('CLAMP_TO_EDGE')))
		elif tex_node.extension=="CLIP":
			samp_res.statements.append(Statement("wrap", Token('CLAMP_TO_BORDER')))
			samp_res.statements.append(Statement("border_color", 0.0, 0.0, 0.0, 0.0))

		return samp_res

	def get_sampler_name(self, tex_node):
		name_parts = []

		use_interpolation = tex_node.interpolation!='Closest'
		name_parts.append("linear" if use_interpolation else "nearest")
		if tex_node.use_mipmap:
			name_parts.append("mip")
		if use_interpolation and tex_node.max_anisotropy>1:
			name_parts.append("aniso{:g}x".format(tex_node.max_anisotropy))
		if tex_node.extension!="REPEAT":
			name_parts.append("clip" if tex_node.extension=="CLIP" else "clamp")

		return "_".join(name_parts)+".samp"
