#ifndef MSP_GL_MODULE_H_
#define MSP_GL_MODULE_H_

#include <string>
#include <msp/io/base.h>
#include "datatype.h"
#include "gl.h"
#include "glsl/compiler.h"
#include "glsl/sourcemap.h"

namespace Msp {
namespace GL {

class invalid_module: public std::runtime_error
{
public:
	invalid_module(const std::string &w): runtime_error(w) { }
	virtual ~invalid_module() throw() { }
};

class Resources;

class Module
{
public:
	enum Format
	{
		GLSL,
		SPIR_V
	};

protected:
	Module() { }
public:
	virtual ~Module() { }

	virtual Format get_format() const = 0;

	void set_source(const std::string &);
	void load_source(IO::Base &, Resources *, const std::string &);
	void load_source(IO::Base &, const std::string &);
private:
	virtual void compile(SL::Compiler &) = 0;
};

class GlslModule: public Module
{
private:
	std::string prepared_source;
	SL::SourceMap source_map;

public:
	virtual Format get_format() const { return GLSL; }

private:
	virtual void compile(SL::Compiler &);

public:
	const std::string &get_prepared_source() const { return prepared_source; }
	const SL::SourceMap &get_source_map() const { return source_map; }
};

class SpirVModule: public Module
{
public:
	enum Stage
	{
		VERTEX = 0,
		GEOMETRY = 3,
		FRAGMENT = 4
	};

	enum StorageClass
	{
		UNIFORM_CONSTANT = 0,
		INPUT = 1,
		UNIFORM = 2,
		OUTPUT = 3
	};

	struct Structure;
	struct Variable;

	struct EntryPoint
	{
		std::string name;
		Stage stage;
		std::vector<const Variable *> globals;

		EntryPoint();
	};

	struct StructMember
	{
		std::string name;
		DataType type;
		const Structure *struct_type;
		unsigned offset;
		unsigned array_size;
		unsigned array_stride;
		unsigned matrix_stride;

		StructMember();
	};

	struct Structure
	{
		std::string name;
		std::vector<StructMember> members;
		unsigned size;
	};

	struct Variable
	{
		std::string name;
		StorageClass storage;
		DataType type;
		const Structure *struct_type;
		unsigned array_size;
		int location;
		int descriptor_set;
		int binding;

		Variable();

		bool operator==(const Variable &) const;
	};

	struct SpecConstant
	{
		std::string name;
		unsigned constant_id;
		DataType type;
	};

private:
	struct TypeInfo
	{
		DataType type;
		const Structure *struct_type;
		unsigned array_size;
		unsigned array_stride;
		StorageClass storage;

		TypeInfo();
	};

	struct Reflection
	{
		typedef std::vector<UInt32>::const_iterator CodeIterator;

		std::map<unsigned, std::string> names;
		std::map<unsigned, Variant> constants;
		std::map<unsigned, TypeInfo> types;
		std::map<unsigned, EntryPoint> entry_points;
		std::map<unsigned, Structure> structs;
		std::map<unsigned, Variable> variables;
		std::map<unsigned, SpecConstant> spec_constants;

		static UInt32 get_opcode(UInt32);
		static CodeIterator get_op_end(const CodeIterator &);
		static std::string read_string(CodeIterator &, const CodeIterator &);

		void reflect_code(const std::vector<UInt32> &);
		void reflect_name(CodeIterator);
		void reflect_member_name(CodeIterator);
		void reflect_entry_point(CodeIterator);
		void reflect_void_type(CodeIterator);
		void reflect_bool_type(CodeIterator);
		void reflect_int_type(CodeIterator);
		void reflect_float_type(CodeIterator);
		void reflect_vector_type(CodeIterator);
		void reflect_matrix_type(CodeIterator);
		void reflect_image_type(CodeIterator);
		void reflect_sampled_image_type(CodeIterator);
		void reflect_array_type(CodeIterator);
		void reflect_struct_type(CodeIterator);
		void reflect_pointer_type(CodeIterator);
		void reflect_constant(CodeIterator);
		void reflect_spec_constant_bool(CodeIterator);
		void reflect_spec_constant(CodeIterator);
		void reflect_variable(CodeIterator);
		void reflect_decorate(CodeIterator);
		void reflect_member_decorate(CodeIterator);
	};

	std::vector<UInt32> code;
	std::vector<EntryPoint> entry_points;
	std::vector<Structure> structs;
	std::vector<Variable> variables;
	std::vector<SpecConstant> spec_constants;

public:
	SpirVModule() { }
	SpirVModule(const SpirVModule &);
	SpirVModule &operator=(const SpirVModule &);
private:
	void remap_pointers_from(const SpirVModule &);

public:
	virtual Format get_format() const { return SPIR_V; }

	void load_code(IO::Base &);
private:
	virtual void compile(SL::Compiler &);

public:
	const std::vector<UInt32> &get_code() const { return code; }
	const std::vector<EntryPoint> &get_entry_points() const { return entry_points; }
	const std::vector<Variable> &get_variables() const { return variables; }
	const std::vector<SpecConstant> &get_spec_constants() const { return spec_constants; }
};

} // namespace GL
} // namespace Msp

#endif
