#ifndef MSP_GL_SL_OPTIMIZE_H_
#define MSP_GL_SL_OPTIMIZE_H_

#include <map>
#include <set>
#include "visitor.h"

namespace Msp {
namespace GL {
namespace SL {

/** Assigns values to specialization constants, turning them into normal
constants. */
class ConstantSpecializer: private TraversingVisitor
{
private:
	const std::map<std::string, int> *values;

public:
	ConstantSpecializer();

	void apply(Stage &, const std::map<std::string, int> &);

private:
	virtual void visit(VariableDeclaration &);
};

/** Finds functions which are candidates for inlining.  Currently this means
functions which have no flow control statements, no more than one return
statement, and are either builtins or only called once. */
class InlineableFunctionLocator: private TraversingVisitor
{
private:
	std::map<FunctionDeclaration *, unsigned> refcounts;
	std::set<FunctionDeclaration *> inlineable;
	FunctionDeclaration *current_function;
	unsigned return_count;

public:
	InlineableFunctionLocator();

	std::set<FunctionDeclaration *> apply(Stage &s) { s.content.visit(*this); return inlineable; }

private:
	virtual void visit(FunctionCall &);
	virtual void visit(FunctionDeclaration &);
	virtual void visit(Conditional &);
	virtual void visit(Iteration &);
	virtual void visit(Return &);
};

/** Injects statements from one function into another.  Local variables are
renamed to avoid conflicts.  After inlining, uses NodeReorderer to cause
dependencies of the inlined statements to appear before the target function. */
class InlineContentInjector: private TraversingVisitor
{
private:
	enum Pass
	{
		REFERENCED,
		INLINE,
		RENAME
	};

	FunctionDeclaration *source_func;
	Block staging_block;
	Pass pass;
	RefPtr<Statement> r_inlined_statement;
	std::set<Node *> dependencies;
	std::set<std::string> referenced_names;
	std::string r_result_name;

public:
	InlineContentInjector();

	std::string apply(Stage &, FunctionDeclaration &, Block &, const NodeList<Statement>::iterator &, FunctionCall &);

private:
	virtual void visit(VariableReference &);
	virtual void visit(InterfaceBlockReference &);
	virtual void visit(FunctionCall &);
	virtual void visit(VariableDeclaration &);
	virtual void visit(Return &);
};

/** Inlines functions.  Internally uses InlineableFunctionLocator to find
candidate functions.  Only functions which consist of a single return statement
are inlined. */
class FunctionInliner: private TraversingVisitor
{
private:
	Stage *stage;
	std::set<FunctionDeclaration *> inlineable;
	FunctionDeclaration *current_function;
	NodeList<Statement>::iterator insert_point;
	RefPtr<Expression> r_inline_result;
	bool r_any_inlined;
	bool r_inlined_here;

public:
	FunctionInliner();

	bool apply(Stage &);

private:
	virtual void visit(RefPtr<Expression> &);
	virtual void visit(Block &);
	virtual void visit(FunctionCall &);
	virtual void visit(FunctionDeclaration &);
	virtual void visit(Iteration &);
};

/** Inlines variables into expressions.  Variables with trivial values (those
consisting of a single literal or variable reference) are always inlined.
Variables which are only referenced once are also inlined. */
class ExpressionInliner: private TraversingVisitor
{
private:
	struct ExpressionInfo
	{
		Expression *expression;
		Block *assign_scope;
		RefPtr<Expression> *inline_point;
		bool trivial;
		bool available;

		ExpressionInfo(): expression(0), assign_scope(0), inline_point(0), trivial(false), available(true) { }
	};

	std::map<Assignment::Target, ExpressionInfo> expressions;
	ExpressionInfo *r_ref_info;
	bool r_any_inlined;
	bool r_trivial;
	bool mutating;
	bool iteration_init;
	Block *iteration_body;
	const Operator *r_oper;

public:
	ExpressionInliner();

	bool apply(Stage &);

private:
	void inline_expression(Expression &, RefPtr<Expression> &);
	virtual void visit(Block &);
	virtual void visit(RefPtr<Expression> &);
	virtual void visit(VariableReference &);
	virtual void visit(MemberAccess &);
	virtual void visit(Swizzle &);
	virtual void visit(UnaryExpression &);
	virtual void visit(BinaryExpression &);
	virtual void visit(Assignment &);
	virtual void visit(TernaryExpression &);
	virtual void visit(FunctionCall &);
	virtual void visit(VariableDeclaration &);
	virtual void visit(Iteration &);
};

/** Replaces expressions consisting entirely of literals with the results of
evaluating the expression.*/
class ConstantFolder: private TraversingVisitor
{
private:
	VariableDeclaration *iteration_var;
	Variant iter_init_value;
	Variant r_constant_value;
	bool iteration_init;
	bool r_constant;
	bool r_literal;
	bool r_uses_iter_var;
	bool r_any_folded;

public:
	bool apply(Stage &s) { s.content.visit(*this); return r_any_folded; }

private:
	static BasicTypeDeclaration::Kind get_value_kind(const Variant &);
	template<typename T>
	static T evaluate_logical(char, T, T);
	template<typename T>
	static bool evaluate_relation(const char *, T, T);
	template<typename T>
	static T evaluate_arithmetic(char, T, T);
	void set_result(const Variant &, bool = false);

	virtual void visit(RefPtr<Expression> &);
	virtual void visit(Literal &);
	virtual void visit(VariableReference &);
	virtual void visit(MemberAccess &);
	virtual void visit(Swizzle &);
	virtual void visit(UnaryExpression &);
	virtual void visit(BinaryExpression &);
	virtual void visit(Assignment &);
	virtual void visit(TernaryExpression &);
	virtual void visit(FunctionCall &);
	virtual void visit(VariableDeclaration &);
	virtual void visit(Iteration &);
};

/** Removes conditional statements and loops where the condition can be
determined as constant at compile time. */
class ConstantConditionEliminator: private TraversingVisitor
{
private:
	enum ConstantStatus
	{
		CONSTANT_FALSE,
		CONSTANT_TRUE,
		NOT_CONSTANT
	};

	NodeList<Statement>::iterator insert_point;
	std::set<Node *> nodes_to_remove;
	RefPtr<Expression> r_ternary_result;

public:
	void apply(Stage &);

private:
	ConstantStatus check_constant_condition(const Expression &);

	virtual void visit(Block &);
	virtual void visit(RefPtr<Expression> &);
	virtual void visit(TernaryExpression &);
	virtual void visit(Conditional &);
	virtual void visit(Iteration &);
};

class UnreachableCodeRemover: private TraversingVisitor
{
private:
	bool reachable;
	std::set<Node *> unreachable_nodes;

public:
	UnreachableCodeRemover();

	virtual bool apply(Stage &);

private:
	virtual void visit(Block &);
	virtual void visit(FunctionDeclaration &);
	virtual void visit(Conditional &);
	virtual void visit(Iteration &);
	virtual void visit(Return &) { reachable = false; }
	virtual void visit(Jump &) { reachable = false; }
};

/** Removes types which are not used anywhere. */
class UnusedTypeRemover: private TraversingVisitor
{
private:
	std::set<Node *> unused_nodes;

public:
	bool apply(Stage &);

private:
	virtual void visit(RefPtr<Expression> &);
	virtual void visit(BasicTypeDeclaration &);
	virtual void visit(ImageTypeDeclaration &);
	virtual void visit(StructDeclaration &);
	virtual void visit(VariableDeclaration &);
	virtual void visit(InterfaceBlock &);
	virtual void visit(FunctionDeclaration &);
};

/** Removes variable declarations with no references to them.  Assignment
statements where the result is not used are also removed. */
class UnusedVariableRemover: private TraversingVisitor
{
private:
	struct AssignmentInfo
	{
		Node *node;
		Assignment::Target target;
		std::vector<Node *> used_by;

		AssignmentInfo(): node(0) { }
	};

	struct VariableInfo
	{
		InterfaceBlock *interface_block;
		std::vector<AssignmentInfo *> assignments;
		bool initialized;
		bool output;
		bool referenced;

		VariableInfo(): interface_block(0), initialized(false), output(false), referenced(false) { }
	};

	typedef std::map<Statement *, VariableInfo> BlockVariableMap;

	Stage *stage;
	BlockVariableMap variables;
	std::list<AssignmentInfo> assignments;
	InterfaceBlock *interface_block;
	Assignment *r_assignment;
	bool assignment_target;
	bool r_side_effects;
	bool in_struct;
	bool composite_reference;
	Assignment::Target r_reference;
	std::set<Node *> unused_nodes;

public:
	UnusedVariableRemover();

	bool apply(Stage &);

private:
	void referenced(const Assignment::Target &, Node &);
	virtual void visit(VariableReference &);
	virtual void visit(InterfaceBlockReference &);
	void visit_composite(Expression &);
	virtual void visit(MemberAccess &);
	virtual void visit(Swizzle &);
	virtual void visit(UnaryExpression &);
	virtual void visit(BinaryExpression &);
	virtual void visit(Assignment &);
	virtual void visit(TernaryExpression &);
	virtual void visit(FunctionCall &);
	void record_assignment(const Assignment::Target &, Node &);
	virtual void visit(ExpressionStatement &);
	virtual void visit(StructDeclaration &);
	virtual void visit(VariableDeclaration &);
	virtual void visit(InterfaceBlock &);
	void merge_variables(const BlockVariableMap &);
	virtual void visit(FunctionDeclaration &);
	virtual void visit(Conditional &);
	virtual void visit(Iteration &);
};

/** Removes function declarations with no references to them. */
class UnusedFunctionRemover: private TraversingVisitor
{
private:
	std::set<Node *> unused_nodes;
	std::set<FunctionDeclaration *> used_definitions;

public:
	bool apply(Stage &s);

private:
	virtual void visit(FunctionCall &);
	virtual void visit(FunctionDeclaration &);
};

} // namespace SL
} // namespace GL
} // namespace Msp

#endif
